#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <math.h>
#include <float.h>

#define TILE_DIM 16
#define WARP_SIZE 32
#define FULL_MASK 0xffffffff

__device__ inline float gelu(float x) {
    return 0.5f * x * (1.0f + erf(x * 0.70710678118654752440f));
}

// Warp-level reduction using shuffle instructions
__device__ inline float warp_reduce_max(float val) {
    for (int offset = WARP_SIZE/2; offset > 0; offset /= 2) {
        float other = __shfl_down_sync(FULL_MASK, val, offset);
        val = fmaxf(val, other);
    }
    return val;
}

__device__ inline float warp_reduce_sum(float val) {
    for (int offset = WARP_SIZE/2; offset > 0; offset /= 2) {
        val += __shfl_down_sync(FULL_MASK, val, offset);
    }
    return val;
}

__global__ void tiled_gemm_kernel(const float* __restrict__ x,
                                   const float* __restrict__ weight,
                                   const float* __restrict__ bias,
                                   float* __restrict__ y,
                                   int batch, int in_features, int out_features) {
    __shared__ float tile_x[TILE_DIM][TILE_DIM];
    __shared__ float tile_w[TILE_DIM][TILE_DIM];

    int row = blockIdx.y * TILE_DIM + threadIdx.y;
    int col = blockIdx.x * TILE_DIM + threadIdx.x;
    float sum = 0.0f;

    for (int t = 0; t < (in_features + TILE_DIM - 1) / TILE_DIM; t++) {
        int idx = t * TILE_DIM + threadIdx.x;
        int idy = t * TILE_DIM + threadIdx.y;

        tile_x[threadIdx.y][threadIdx.x] = (row < batch && idx < in_features) ?
            x[row * in_features + idx] : 0.0f;
        tile_w[threadIdx.y][threadIdx.x] = (col < out_features && idy < in_features) ?
            weight[col * in_features + idy] : 0.0f;
        __syncthreads();

        #pragma unroll
        for (int k = 0; k < TILE_DIM; k++) {
            sum += tile_x[threadIdx.y][k] * tile_w[k][threadIdx.x];
        }
        __syncthreads();
    }

    if (row < batch && col < out_features) {
        y[row * out_features + col] = sum + bias[col];
    }
}

__global__ void fused_warp_reduce_gelu_kernel(float* data,
                                               int batch,
                                               int out_features,
                                               int max_dim) {
    extern __shared__ float smem[];
    const int tid = threadIdx.x;
    const int lane_id = tid % WARP_SIZE;
    const int warp_id = tid / WARP_SIZE;
    const int warps_per_block = blockDim.x / WARP_SIZE;

    float thread_max = -FLT_MAX;
    float thread_sum = 0.0f;

    if (max_dim == 0) {
        // Reduce along batch dimension
        const int col = blockIdx.x;
        
        // Each thread processes multiple elements with grid stride
        for (int i = tid; i < batch; i += blockDim.x) {
            float val = data[i * out_features + col];
            thread_max = fmaxf(thread_max, val);
        }

        // Warp-level reduction
        float warp_max = warp_reduce_max(thread_max);

        // Store warp results in shared memory
        if (lane_id == 0) {
            smem[warp_id] = warp_max;
        }
        __syncthreads();

        // Final reduction across warps
        if (warp_id == 0) {
            float block_max = (tid < warps_per_block) ? smem[tid] : -FLT_MAX;
            block_max = warp_reduce_max(block_max);

            // First thread has the final result
            if (tid == 0) {
                smem[0] = block_max;
            }
        }
        __syncthreads();

        float max_val = smem[0];
        
        // Compute mean using the same pattern
        for (int i = tid; i < batch; i += blockDim.x) {
            int idx = i * out_features + col;
            float val = data[idx] - max_val;
            thread_sum += val;
        }

        float warp_sum = warp_reduce_sum(thread_sum);
        if (lane_id == 0) {
            smem[warp_id] = warp_sum;
        }
        __syncthreads();

        if (warp_id == 0) {
            float block_sum = (tid < warps_per_block) ? smem[tid] : 0.0f;
            block_sum = warp_reduce_sum(block_sum);
            if (tid == 0) {
                smem[0] = block_sum / batch;
            }
        }
        __syncthreads();

        float mean = smem[0];

        // Apply GELU with grid stride loop
        for (int i = tid; i < batch; i += blockDim.x) {
            int idx = i * out_features + col;
            float val = data[idx] - mean;
            data[idx] = gelu(val);
        }
    } else {
        // Reduce along feature dimension
        const int row = blockIdx.x;
        
        for (int j = tid; j < out_features; j += blockDim.x) {
            float val = data[row * out_features + j];
            thread_max = fmaxf(thread_max, val);
        }

        float warp_max = warp_reduce_max(thread_max);
        if (lane_id == 0) {
            smem[warp_id] = warp_max;
        }
        __syncthreads();

        if (warp_id == 0) {
            float block_max = (tid < warps_per_block) ? smem[tid] : -FLT_MAX;
            block_max = warp_reduce_max(block_max);
            if (tid == 0) {
                smem[0] = block_max;
            }
        }
        __syncthreads();

        float max_val = smem[0];

        for (int j = tid; j < out_features; j += blockDim.x) {
            int idx = row * out_features + j;
            float val = data[idx] - max_val;
            thread_sum += val;
        }

        float warp_sum = warp_reduce_sum(thread_sum);
        if (lane_id == 0) {
            smem[warp_id] = warp_sum;
        }
        __syncthreads();

        if (warp_id == 0) {
            float block_sum = (tid < warps_per_block) ? smem[tid] : 0.0f;
            block_sum = warp_reduce_sum(block_sum);
            if (tid == 0) {
                smem[0] = block_sum / out_features;
            }
        }
        __syncthreads();

        float mean = smem[0];

        for (int j = tid; j < out_features; j += blockDim.x) {
            int idx = row * out_features + j;
            float val = data[idx] - mean;
            data[idx] = gelu(val);
        }
    }
}

torch::Tensor forward(torch::Tensor x, int max_dim, torch::Tensor weight, torch::Tensor bias) {
    int batch = x.size(0);
    int in_features = x.size(1);
    int out_features = weight.size(0);

    auto y = torch::empty({batch, out_features}, x.options());

    dim3 blockDimGEMM(TILE_DIM, TILE_DIM);
    dim3 gridDimGEMM((out_features + TILE_DIM - 1) / TILE_DIM,
                     (batch + TILE_DIM - 1) / TILE_DIM);
    
    tiled_gemm_kernel<<<gridDimGEMM, blockDimGEMM>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        y.data_ptr<float>(),
        batch, in_features, out_features
    );

    auto max_out = torch::empty({max_dim == 0 ? 1 : batch, max_dim == 0 ? out_features : 1}, y.options());

    int threads = 256; // Must be multiple of WARP_SIZE
    int blocks = max_dim == 0 ? out_features : batch;
    int sharedMemSize = (threads / WARP_SIZE) * sizeof(float);

    fused_warp_reduce_gelu_kernel<<<blocks, threads, sharedMemSize>>>(
        max_out.data_ptr<float>(),
        batch,
        out_features,
        max_dim
    );

    return max_out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Warp-optimized GEMM and reduction with GELU");
}